from collections.abc import Generator
from typing import Protocol

import numpy as np


class DataLoader(Protocol):
    n_val: int
    n_samples: int

    def __init__(self, batch_size: int = 32, validation_split: float = 0.0) -> None: ...

    def train_generator(
        self,
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]: ...

    def val_generator(
        self,
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]: ...



import numpy as np

from toynet.config import NUMERICAL_EPS
from toynet.functions.abstract import ActivationFunction


class ReLU(ActivationFunction):
    def __init__(self) -> None:
        self._fn = lambda x: np.maximum(0, x)
        self._fn_derivative = lambda x: np.where(x > 0, 1, 0)


class Sigmoid(ActivationFunction):
    def __init__(self) -> None:
        self._fn = lambda x: np.clip(1 / (1 + np.exp(-x)), NUMERICAL_EPS, 1 - NUMERICAL_EPS)
        self._fn_derivative = lambda x: self._fn(x) * (1 - self._fn(x))


class Identity(ActivationFunction):
    def __init__(self) -> None:
        self._fn = lambda x: x
        self._fn_derivative = lambda x: np.ones_like(x)


class Softmax(ActivationFunction):
    def __init__(self) -> None:
        self._fn = self._num_safe_softmax
        self._fn_derivative = self._num_safe_softmax_deriv

    def _num_safe_softmax(self, x: np.ndarray) -> np.ndarray:
        shifted = x - np.max(x, axis=1, keepdims=True)
        exp_vals = np.exp(shifted)
        return exp_vals / np.sum(exp_vals, axis=1, keepdims=True)  # type: ignore[no-any-return]

    def _num_safe_softmax_deriv(self, x: np.ndarray) -> np.ndarray:
        s = self._num_safe_softmax(x)
        return np.ones_like(s)



from toynet.networks.protocol import NeuralNetwork
from toynet.policies.protocol import Policy


class PeriodicallySaveModel(Policy):
    def __init__(self, save_path: str, save_interval: int) -> None:
        self.network: NeuralNetwork | None = None
        self.save_path = save_path
        self.save_interval = save_interval
        self.current_step = 0

    def set_network(self, network: NeuralNetwork) -> None:
        self.network = network

    def __call__(self, _: float) -> bool:
        if self.current_step % self.save_interval == 0 and self.network:
            self.network.logger.info("Saving model to <%s>", self.save_path)
            self.network.save_to_npz(self.save_path)
        self.current_step += 1
        return False

    def __str__(self) -> str:
        return str(self.__class__.__name__)



from abc import ABC, abstractmethod
from collections.abc import Callable

import numpy as np


class ActivationFunction(ABC):
    _fn: Callable[[np.ndarray], np.ndarray]
    _fn_derivative: Callable[[np.ndarray], np.ndarray]

    @abstractmethod
    def __init__(self) -> None: ...

    def __call__(self, x: np.ndarray, derivative: bool = False) -> np.ndarray:
        if derivative:
            return self._fn_derivative(x)
        return self._fn(x)


class LossFunction(ABC):
    _fn: Callable[[np.ndarray, np.ndarray], np.ndarray]
    _fn_derivative: Callable[[np.ndarray, np.ndarray], np.ndarray]

    @abstractmethod
    def __init__(self) -> None: ...

    def __call__(
        self, y_true: np.ndarray, y_pred: np.ndarray, derivative: bool = False
    ) -> np.ndarray:
        if derivative:
            return self._fn_derivative(y_true, y_pred)
        return self._fn(y_true, y_pred)



from typing import Protocol

import numpy as np

from toynet.functions.abstract import ActivationFunction
from toynet.initalizers.abstract import Initializer


class Layer(Protocol):
    activation_fn: ActivationFunction
    z: np.ndarray
    h: np.ndarray
    last_input: np.ndarray

    def __init__(
        self,
        in_values_count: int,
        out_values_count: int,
        activation_fn: ActivationFunction,
        initializer: Initializer,
    ) -> None: ...

    def forward(self, layer_input: np.ndarray, apply_activation: bool = True) -> np.ndarray: ...

    def backpass(self, grad_input: np.ndarray) -> np.ndarray: ...

    def reset_accumulated_gradient(self) -> None: ...

    @property
    def parameters(self) -> list[np.ndarray]: ...

    @property
    def gradients(self) -> list[np.ndarray]: ...



from toynet.networks.protocol import NeuralNetwork
from toynet.policies.protocol import Policy


class ReduceLROnPlateau(Policy):
    def __init__(self, factor: float = 0.1, patience: int = 10, min_lr: float = 1e-6) -> None:
        self.network: NeuralNetwork | None = None
        self.factor = factor
        self.patience = patience
        self.min_lr = min_lr
        self.waiting = 0
        self.best_validation_loss = float("inf")

    def set_network(self, network: NeuralNetwork) -> None:
        self.network = network

    def __call__(self, validation_loss: float) -> bool:
        if not self.network:
            return False

        if validation_loss < self.best_validation_loss:
            self.best_validation_loss = validation_loss
            self.waiting = 0
            return False
        self.waiting += 1
        if self.waiting >= self.patience:
            new_lr = self.network.optimizer.learning_rate * self.factor
            self.network.logger.info(
                "Reducing learning rate from %s to %s", self.network.optimizer.learning_rate, new_lr
            )
            if new_lr < self.min_lr:
                return True
            self.network.optimizer.learning_rate = new_lr
            self.waiting = 0
        return False

    def __str__(self) -> str:
        return str(self.__class__.__name__) + f"(min_lr={self.min_lr})"



import numpy as np

from toynet.optimizers.protocol import Optimizer


class BasicOptimizer(Optimizer):
    def __init__(self, learning_rate: float = 0.01) -> None:
        self.learning_rate = learning_rate

    def step(self, params: list[np.ndarray], grads: list) -> None:
        params[0][...] -= self.learning_rate * grads[0]
        params[1][...] -= self.learning_rate * grads[1]



import logging

import numpy as np

from toynet.config import get_simple_logger
from toynet.data_loaders import DataLoader
from toynet.functions import LossFunction
from toynet.layers import Layer
from toynet.optimizers import Optimizer
from toynet.policies.protocol import Policy


class MultiLayerPerceptron:
    def __init__(
        self,
        layers: list[Layer],
        loss_function: type[LossFunction],
        optimizer: Optimizer,
        logger: logging.Logger | None = None,
    ) -> None:
        self._layers: list[Layer] = layers
        self.loss_function: LossFunction = loss_function()
        self.optimizer: Optimizer = optimizer

        if logger is None:
            self.logger = get_simple_logger(__name__)
        else:
            self.logger = logger

    def forward(self, network_input: np.ndarray) -> np.ndarray:
        curr_vector = network_input.copy()
        for layer in self._layers:
            curr_vector = layer.forward(curr_vector)
        return curr_vector

    def backprop(self, network_input: np.ndarray, y_true: np.ndarray) -> np.ndarray:
        y_pred = self.forward(network_input)

        grad = self.loss_function(y_true, y_pred, derivative=True)
        for layer in reversed(self._layers):
            grad = layer.backpass(grad)
        return grad

    def train(  # noqa: C901
        self, data_loader: DataLoader, epochs: int, policies: list[Policy] | None = None
    ) -> None:
        if policies is None:
            policies = []
        for policy in policies:
            policy.set_network(self)

        self.logger.info("Starting training...")
        for epoch in range(epochs):
            total_loss = 0.0
            for batch_x, batch_y in data_loader.train_generator():
                for layer in self._layers:
                    layer.reset_accumulated_gradient()

                batch_loss = 0.0

                for i in range(len(batch_x)):
                    out = self.forward(batch_x[i])
                    batch_loss += float(self.loss_function(batch_y[i], out)[0][0])
                    self.backprop(batch_x[i], batch_y[i])

                for layer in self._layers:
                    averaged_grads = [grad / len(batch_x) for grad in layer.gradients]
                    self.optimizer.step(layer.parameters, averaged_grads)

                total_loss += batch_loss
            self.logger.info(
                "Epoch %d/%d - Loss: %.4f", epoch + 1, epochs, total_loss / data_loader.n_samples
            )
            if data_loader.n_val > 0:
                val_loss = 0.0
                for X_val, y_val in data_loader.val_generator():
                    for i in range(len(X_val)):
                        out = self.forward(X_val[i])
                        val_loss += float(self.loss_function(y_val[i], out)[0][0])
                avg_val_loss = val_loss / data_loader.n_val
                for policy in policies:
                    if policy(avg_val_loss):
                        self.logger.info(
                            "Stopping training at epoch %d due to policy %s", epoch + 1, policy
                        )
                        return
                self.logger.info(
                    "Epoch %d/%d - Validation Loss: %.4f",
                    epoch + 1,
                    epochs,
                    avg_val_loss,
                )

    def save_to_npz(self, filename: str) -> None:
        data = {}
        for i, layer in enumerate(self._layers):
            weights, bias = layer.parameters
            data[f"W{i}"] = weights
            data[f"b{i}"] = bias
        np.savez(filename, **data)  # type: ignore[arg-type]

    def load_from_npz(self, filename: str) -> None:
        archive = np.load(filename, allow_pickle=True)
        for i, layer in enumerate(self._layers):
            layer._weights = archive[f"W{i}"]  # type: ignore[attr-defined] # noqa: SLF001
            layer._bias = archive[f"b{i}"]  # type: ignore[attr-defined] # noqa: SLF001

    def __call__(self, network_input: np.ndarray) -> np.ndarray:
        return self.forward(network_input)



import logging

NUMERICAL_EPS = 1e-12


def get_simple_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)

    if logger.handlers:
        return logger

    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            "%(asctime)s [%(name)s] [%(levelname)s] - %(message)s", datefmt="%H:%M:%S"
        )
    )
    logger.addHandler(handler)

    logger.propagate = False
    return logger



import numpy as np

from toynet.optimizers.protocol import Optimizer


class Adam(Optimizer):
    def __init__(
        self,
        learning_rate: float = 0.001,
        beta1: float = 0.9,
        beta2: float = 0.999,
        eps: float = 1e-8,
    ) -> None:
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.t = 0

        self.m: dict[int, np.ndarray] = {}
        self.v: dict[int, np.ndarray] = {}

    def step(self, params: list[np.ndarray], grads: list) -> None:
        self.t += 1
        for p, g in zip(params, grads, strict=False):
            if g is None:
                continue

            key = id(p)

            m: np.ndarray = self.m.get(key, np.zeros_like(g))
            v: np.ndarray = self.v.get(key, np.zeros_like(g))

            m = self.beta1 * m + (1.0 - self.beta1) * g
            v = self.beta2 * v + (1.0 - self.beta2) * (g**2)

            m_hat = m / (1.0 - self.beta1**self.t)
            v_hat = v / (1.0 - self.beta2**self.t)

            update = (self.learning_rate * m_hat) / (np.sqrt(v_hat) + self.eps)

            p[...] -= update

            self.m[key] = m
            self.v[key] = v



import logging
from typing import Protocol

import numpy as np

from toynet.data_loaders.protocol import DataLoader
from toynet.functions.abstract import LossFunction
from toynet.layers.protocol import Layer
from toynet.optimizers.protocol import Optimizer
from toynet.policies.protocol import Policy


class NeuralNetwork(Protocol):
    optimizer: Optimizer
    logger: logging.Logger

    def __init__(
        self, layers: list[Layer], loss_function: LossFunction, optimizer: Optimizer
    ) -> None: ...

    def forward(self, network_input: np.ndarray) -> np.ndarray: ...

    def backprop(self, network_input: np.ndarray, y_true: np.ndarray) -> np.ndarray: ...

    def train(
        self, data_loader: DataLoader, epochs: int, policies: list[Policy] | None = None
    ) -> None: ...

    def save_to_npz(self, filename: str) -> None: ...

    def load_from_npz(self, filename: str) -> None: ...

    def __call__(self, network_input: np.ndarray) -> np.ndarray: ...



from toynet.data_loaders import BasicDataLoader, CSVDataLoader
from toynet.functions import (
    BinaryCrossEntropy,
    CategoricalCrossEntropy,
    Identity,
    MeanSquaredError,
    ReLU,
    Sigmoid,
    Softmax,
)
from toynet.layers import Dense
from toynet.networks.mlp import MultiLayerPerceptron
from toynet.optimizers import Adam, BasicOptimizer
from toynet.policies import (
    PeriodicallySaveModel,
    ReduceLROnPlateau,
    SaveBestModel,
    ValidationLossEarlyStop,
)

__all__ = [
    "Adam",
    "BasicDataLoader",
    "BasicOptimizer",
    "BinaryCrossEntropy",
    "CSVDataLoader",
    "CategoricalCrossEntropy",
    "Dense",
    "Identity",
    "MeanSquaredError",
    "MultiLayerPerceptron",
    "PeriodicallySaveModel",
    "ReLU",
    "ReduceLROnPlateau",
    "SaveBestModel",
    "Sigmoid",
    "Softmax",
    "ValidationLossEarlyStop",
]



from toynet.networks.protocol import NeuralNetwork
from toynet.policies.protocol import Policy


class ValidationLossEarlyStop(Policy):
    def __init__(self, save_path: str, patience: int, should_save: bool = True) -> None:
        self.network: NeuralNetwork | None = None
        self.save_path = save_path
        self.patience = patience
        self.best_loss = float("inf")
        self.should_save = should_save
        self.counter = 0

    def set_network(self, network: NeuralNetwork) -> None:
        self.network = network

    def _should_stop(self, current_loss: float) -> bool:
        if current_loss < self.best_loss:
            self.best_loss = current_loss
            self.counter = 0
            return False
        self.counter += 1
        return self.counter >= self.patience

    def __call__(self, validation_loss: float) -> bool:
        if self._should_stop(validation_loss):
            if self.should_save and self.network:
                self.network.save_to_npz(self.save_path)
            return True
        return False

    def __str__(self) -> str:
        return str(self.__class__.__name__) + f"(patience={self.patience})"



from toynet.data_loaders.basic_data_loader import BasicDataLoader
from toynet.data_loaders.csv_data_loader import CSVDataLoader
from toynet.data_loaders.protocol import DataLoader

__all__ = [
    "BasicDataLoader",
    "CSVDataLoader",
    "DataLoader",
]



import numpy as np

from toynet.config import NUMERICAL_EPS
from toynet.functions.abstract import LossFunction


class BinaryCrossEntropy(LossFunction):
    def __init__(self) -> None:
        self._fn = self._num_safe_bce
        self._fn_derivative = self._num_safe_bce_deriv

    def _num_safe_bce(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        y_pred_clipped = np.clip(y_pred, NUMERICAL_EPS, 1 - NUMERICAL_EPS)
        return -(y_true * np.log(y_pred_clipped) + (1 - y_true) * np.log(1 - y_pred_clipped))  # type: ignore[no-any-return]

    def _num_safe_bce_deriv(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        y_pred = np.clip(y_pred, NUMERICAL_EPS, 1 - NUMERICAL_EPS)
        return -(y_true / y_pred) + (1 - y_true) / (1 - y_pred)  # type: ignore[no-any-return]


class MeanSquaredError(LossFunction):
    def __init__(self) -> None:
        self._fn = lambda y_true, y_pred: np.array([np.mean((y_true - y_pred) ** 2)]).reshape(-1, 1)
        self._fn_derivative = lambda y_true, y_pred: 2 * (y_pred - y_true) / y_true.size


class CategoricalCrossEntropy(LossFunction):
    def __init__(self) -> None:
        self._fn = self._num_safe_cce
        self._fn_derivative = self._num_safe_cce_derivative

    def _num_safe_cce(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        y_pred_clipped = np.clip(y_pred, NUMERICAL_EPS, 1 - NUMERICAL_EPS)
        return -np.sum(y_true * np.log(y_pred_clipped), axis=-1).reshape(-1, 1)  # type: ignore[no-any-return]

    def _num_safe_cce_derivative(self, y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        return y_pred - y_true  # type: ignore[no-any-return]



from toynet.policies.periodically_save_model import PeriodicallySaveModel
from toynet.policies.protocol import Policy
from toynet.policies.reduce_lr_on_plateu import ReduceLROnPlateau
from toynet.policies.save_best_model import SaveBestModel
from toynet.policies.validation_early_stop import ValidationLossEarlyStop

__all__ = [
    "PeriodicallySaveModel",
    "Policy",
    "ReduceLROnPlateau",
    "SaveBestModel",
    "ValidationLossEarlyStop",
]



import numpy as np

from toynet.functions.abstract import ActivationFunction
from toynet.initalizers.abstract import Initializer
from toynet.initalizers.he import He
from toynet.layers.protocol import Layer


class Dense(Layer):
    def __init__(
        self,
        in_values_count: int,
        out_values_count: int,
        activation_fn: type[ActivationFunction],
        initializer: type[Initializer] = He,
    ) -> None:
        self._weights, self._bias = initializer()(in_values_count, out_values_count)

        self.grad_w_accum = np.zeros_like(self._weights)
        self.grad_b_accum = np.zeros_like(self._bias)

        self.activation_fn = activation_fn()
        self.z: np.ndarray = np.zeros((1, out_values_count))
        self.h: np.ndarray = np.zeros((1, out_values_count))
        self.last_input: np.ndarray = np.zeros((in_values_count, 1))

    def forward(self, layer_input: np.ndarray, apply_activation: bool = True) -> np.ndarray:
        F = lambda x: self.activation_fn(x) if apply_activation else x  # noqa: E731
        self.last_input = layer_input
        self.z = layer_input @ self._weights + self._bias
        self.h = F(self.z)
        return self.h

    def backpass(
        self,
        grad_input: np.ndarray,
    ) -> np.ndarray:
        grad_z = grad_input * self.activation_fn(self.z, derivative=True)
        grad_w = self.last_input.T @ grad_z
        grad_output = grad_z @ self._weights.T

        self.grad_w_accum += grad_w
        self.grad_b_accum += grad_z

        return grad_output  # type: ignore[no-any-return]

    def reset_accumulated_gradient(self) -> None:
        self.grad_w_accum.fill(0)
        self.grad_b_accum.fill(0)

    @property
    def parameters(self) -> list[np.ndarray]:
        return [self._weights, self._bias]

    @property
    def gradients(self) -> list[np.ndarray]:
        return [self.grad_w_accum, self.grad_b_accum]



from collections.abc import Generator

import numpy as np

from toynet.data_loaders.protocol import DataLoader


class BasicDataLoader(DataLoader):
    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        batch_size: int = 32,
        validation_split: float = 0.0,
    ) -> None:
        self.X = X
        self.y = y
        self.batch_size = batch_size
        self.validation_split = validation_split

        self.n_samples = len(X)
        self.n_val = int(self.n_samples * validation_split)
        self.n_train = self.n_samples - self.n_val

    def train_generator(
        self,
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]:
        X = []
        y = []
        for i in range(self.n_train):
            X.append(self.X[i].reshape(1, -1))
            y.append(self.y[i])
        yield np.array(X), np.array(y)

    def val_generator(
        self,
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]:
        X = []
        y = []
        for i in range(self.n_train, self.n_samples):
            X.append(self.X[i].reshape(1, -1))
            y.append(self.y[i])
        yield np.array(X), np.array(y)



from typing import Protocol

import numpy as np


class Optimizer(Protocol):
    learning_rate: float

    def __init__(self, learning_rate: float = 0.01) -> None: ...

    def step(self, params: list[np.ndarray], grads: list) -> None: ...



from toynet.networks.protocol import NeuralNetwork
from toynet.policies.protocol import Policy


class SaveBestModel(Policy):
    def __init__(self, save_path: str, save_grace_period: int) -> None:
        self.network: NeuralNetwork | None = None
        self.save_path = save_path
        self.best_validation_loss = float("inf")
        self.grace_period = save_grace_period
        self.waiting = 0

    def set_network(self, network: NeuralNetwork) -> None:
        self.network = network

    def __call__(self, validation_loss: float) -> bool:
        if self.waiting < self.grace_period:
            self.waiting += 1
            return False

        if validation_loss < self.best_validation_loss:
            self.best_validation_loss = validation_loss
            if self.network:
                self.network.save_to_npz(self.save_path)
        return False

    def __str__(self) -> str:
        return str(self.__class__.__name__)



from collections.abc import Callable, Generator
from pathlib import Path

import numpy as np
import pandas as pd

from toynet.data_loaders.protocol import DataLoader


class CSVDataLoader(DataLoader):
    def __init__(  # noqa: PLR0913
        self,
        csv_file: str,
        label_cols: list[str],
        batch_size: int = 32,
        validation_split: float = 0.0,
        shuffle: bool = True,
        transform: Callable[[np.ndarray, np.ndarray], tuple[np.ndarray, np.ndarray]] | None = None,
    ) -> None:
        self.csv_file = csv_file
        self.batch_size = batch_size
        self.label_cols = label_cols
        self.shuffle = shuffle
        self.transform = transform

        self.columns = pd.read_csv(csv_file, nrows=0).columns.tolist()

        with Path.open(Path(csv_file), mode="rb") as f:
            self.n_samples = sum(1 for _ in f) - 1
        self.n_val = int(self.n_samples * validation_split)
        self.n_train = self.n_samples - self.n_val

    def _get_reader(self, start_idx: int = 0, end_idx: int | None = None) -> pd.DataFrame:
        nrows = None if end_idx is None else end_idx - start_idx
        skiprows = range(1, start_idx + 1)
        return pd.read_csv(self.csv_file, skiprows=skiprows, nrows=nrows)

    def _batch_generator(
        self, start_idx: int, end_idx: int
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]:
        reader = self._get_reader(start_idx, end_idx)
        if self.shuffle:
            reader = reader.sample(frac=1).reset_index(drop=True)

        y = reader[self.label_cols].to_numpy(dtype=np.float32)
        X = []
        for _, row in reader.iterrows():
            X.append(row.drop(labels=self.label_cols).to_numpy(dtype=np.float32).reshape(1, -1))

        X_array = np.array(X)
        if self.transform:
            X_array, y = self.transform(X_array, y)

        yield X_array, y

    def train_generator(
        self,
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]:
        return self._batch_generator(0, self.n_train)

    def val_generator(
        self,
    ) -> Generator[tuple[np.ndarray, np.ndarray], None, None]:
        if self.n_val == 0:
            return iter(())  # type: ignore[return-value]
        return self._batch_generator(self.n_train, self.n_samples)



from abc import ABC, abstractmethod

import numpy as np


class Initializer(ABC):
    @abstractmethod
    def initialize_weights(self, in_amount: int, out_amount: int) -> np.ndarray: ...

    @abstractmethod
    def initialize_bias(self, out_amount: int) -> np.ndarray: ...

    def __call__(self, in_amount: int, out_amount: int) -> tuple[np.ndarray, np.ndarray]:
        weights = self.initialize_weights(in_amount, out_amount)
        bias = self.initialize_bias(out_amount)
        return weights, bias



from toynet.optimizers.adam import Adam
from toynet.optimizers.basic_optimizer import BasicOptimizer
from toynet.optimizers.protocol import Optimizer

__all__ = ["Adam", "BasicOptimizer", "Optimizer"]



from toynet.functions.abstract import ActivationFunction, LossFunction
from toynet.functions.activation import Identity, ReLU, Sigmoid, Softmax
from toynet.functions.loss import (
    BinaryCrossEntropy,
    CategoricalCrossEntropy,
    MeanSquaredError,
)

__all__ = [
    "ActivationFunction",
    "BinaryCrossEntropy",
    "CategoricalCrossEntropy",
    "Identity",
    "LossFunction",
    "MeanSquaredError",
    "ReLU",
    "Sigmoid",
    "Softmax",
]



import numpy as np

from toynet.initalizers.abstract import Initializer


class Xavier(Initializer):
    def initialize_weights(self, in_amount: int, out_amount: int) -> np.ndarray:
        xavier_std = (2.0 / (in_amount + out_amount)) ** 0.5
        return np.random.normal(0.0, xavier_std, (in_amount, out_amount))

    def initialize_bias(self, out_amount: int) -> np.ndarray:
        return np.zeros((1, out_amount))
